import typer

from pathlib import Path
from rich import print as rprint
from typing_extensions import Annotated


def register_workspace_initialize(app: typer.Typer):
    @app.command(name="initialize")
    def workspace_initialize(
        workspace_name: str,
        out_path: Path | None = None,
        force: Annotated[
            bool, typer.Option("--force", help="Overwrite existing workspace")
        ] = False,
    ) -> None:
        """Create a folder to store data related to a workspace."""
        from ow.workspace import Workspace

        try:
            workspace = Workspace(
                name=workspace_name,
                out_path=out_path,
            )
            workspace_config = workspace.create_workspace(
                out_path,
                force,
            )
            rprint(f"✅ Workspace initialized at: {workspace_config.workspace_path}")
        except:
            rprint("⚠️  [yellow]Unable to create workspace directory[/yellow]")
            _ = typer.Exit()

    _ = app.command(name="init")(workspace_initialize)

    return workspace_initialize



from mcp.server import FastMCP

from typing import Union


def register_workspace_initialize(app: FastMCP):
    from ow.mcp.types import ToolSuccess, ToolError
    from ow.mcp.utils import tool_success, tool_error
    from ow.workspace.config import WorkspaceConfig

    @app.tool(
        title="Initialize Workspace",
        description="Creates new workspace folder for storing outputs.",
        structured_output=True,
    )
    def workspace_initialize(
        workspace_name: str,
        force: bool = False,
    ) -> Union[ToolSuccess[WorkspaceConfig], ToolError]:
        """Create a folder to store data related to a workspace."""
        from ow.workspace import Workspace

        try:
            workspace = Workspace(name=workspace_name)
            workspace_config = workspace.create_workspace(force=force)
            return tool_success(workspace_config)

        except PermissionError as e:
            return tool_error(
                "Permission denied when creating workspace",
                "PERMISSION_DENIED",
                workspace_name=workspace_name,
                exception_type=type(e).__name__,
            )

        except FileExistsError as e:
            return tool_error(
                "Workspace already exists, use `force` to overwrite existing workspace",
                "WORKSPACE_EXISTS",
                workspace_name=workspace_name,
                suggestion="Use force=True to overwrite",
                exception_message=str(e),
            )

        except Exception as e:
            return tool_error(
                "Failed to create workspace",
                "WORKSPACE_CREATE_FAILED",
                workspace_name=workspace_name,
                exception_type=type(e).__name__,
                exception_message=str(e),
            )

    _ = workspace_initialize



from enum import Enum

from mcp.server import FastMCP


class Subfolder(str, Enum):
    meshes = "meshes"
    parts = "parts"
    segments = "segments"


def register_workspace_list(app: FastMCP):
    from ow.mcp.types import ToolSuccess
    from ow.mcp.utils import tool_success

    @app.tool(
        title="List Workspaces",
        description="Provides a list of created workspaces.",
        structured_output=True,
    )
    def workspaces() -> ToolSuccess[list[str] | None]:
        from ow.workspace.list import list_workspaces

        return tool_success(list_workspaces())

    # @app.tool(
    #     title="List Workspace Folder",
    #     description="Provides a list of specific output folders created by solver within specified workspace. This includes `meshes`, `parts`, `segments`",
    #     structured_output=True,
    # )
    # def workspace_subfolder_list(
    #    workspace: str,
    #    subfolder: Subfolder
    # ) -> Union[ToolSuccess[list[str] | None], ToolError]:
    #     """
    #     Lists available meshes within workspace
    #     """
    #     from am.cli.utils import get_workspace_path
    #
    #     workspace_path = get_workspace_path(workspace)
    #
    #     outputs = None
    #     match subfolder:
    #         case Subfolder.meshes:
    #             from am.workspace.list import list_workspace_meshes
    #             outputs = list_workspace_meshes(workspace_path)
    #         case Subfolder.parts:
    #             from am.workspace.list import list_workspace_parts
    #             outputs = list_workspace_parts(workspace_path)
    #         case Subfolder.segments:
    #             from am.workspace.list import list_workspace_segments
    #             outputs = list_workspace_segments(workspace_path)
    #         case _:
    #             return tool_error(
    #                 "Output folder not valid",
    #                 "INVALID_OUTPUT_FOLDER",
    #                 workspace_name=workspace,
    #             )
    #
    #     return tool_success(outputs)

    # _ = ( workspaces, workspace_subfolder_list)
    _ = workspaces


def register_workspace_list_resources(app: FastMCP):

    @app.resource("workspace://")
    def workspace_list() -> list[str] | None:
        from ow.workspace.list import list_workspaces

        return list_workspaces()

    _ = workspace_list

    # @app.resource("workspace://{workspace}/meshes")
    # def workspace_meshes_list(workspace: str) -> list[str] | None:
    #     """
    #     Lists available meshes within workspace
    #     """
    #     from am.workspace.list import list_workspace_meshes
    #     from am.cli.utils import get_workspace_path
    #
    #     workspace_path = get_workspace_path(workspace)
    #     workspace_meshes = list_workspace_meshes(workspace_path)
    #
    #     return workspace_meshes
    #
    # @app.resource("workspace://{workspace}/part")
    # def workspace_part_list(workspace: str) -> list[str] | None:
    #     """
    #     Lists available parts within workspace
    #     """
    #     from am.workspace.list import list_workspace_parts
    #     from am.cli.utils import get_workspace_path
    #
    #     workspace_path = get_workspace_path(workspace)
    #     workspace_parts = list_workspace_parts(workspace_path)
    #
    #     return workspace_parts
    #
    # @app.resource("workspace://{workspace}/segments")
    # def workspace_segments_list(workspace: str) -> list[str] | None:
    #     """
    #     Lists available segments within workspace
    #     """
    #     from am.workspace.list import list_workspace_segments
    #     from am.cli.utils import get_workspace_path
    #
    #     workspace_path = get_workspace_path(workspace)
    #     workspace_segments = list_workspace_segments(workspace_path)
    #
    #     return workspace_segments
    #
    # _ = (
    #         workspace_list,
    #         workspace_meshes_list,
    #         workspace_part_list,
    #         workspace_segments_list
    #     )



import os

from pathlib import Path

from ow.workspace.config import WorkspaceConfig


def list_workspaces(out_path: Path | None = None) -> list[str] | None:
    """
    Lists workspace directories within out_path
    """
    if out_path is None:
        project_root = WorkspaceConfig.get_project_root_from_package()
        out_path = project_root / "out"

    if not out_path.exists():
        os.makedirs(out_path)

    if not out_path.is_dir():
        raise FileNotFoundError

    return [
        workspace_dir.name
        for workspace_dir in out_path.iterdir()
        if workspace_dir.is_dir()
    ]


# def list_workspace_meshes(workspace_path: Path) -> list[str] | None:
#     """
#     Lists meshes within workspace directory
#     """
#     meshes_path = workspace_path / "meshes"
#
#     meshes = []
#     for item in meshes_path.iterdir():
#         if item.is_dir() and (item / "timesteps").is_dir():
#             meshes.append(item.name)
#
#     return meshes
#
# def list_workspace_parts(
#         workspace_path: Path,
#         suffix: str = ".gcode",
# ) -> list[str] | None:
#     """
#     Lists parts within workspace directory
#     """
#     parts_path = workspace_path / "parts"
#
#     return [
#         partfile.name
#         for partfile in parts_path.iterdir()
#         if partfile.is_file() and partfile.suffix == suffix
#     ]
#
#
# def list_workspace_segments(workspace_path: Path) -> list[str] | None:
#     """
#     Lists segments within workspace directory
#     """
#     segments_path = workspace_path / "segments"
#
#     segments = []
#     for item in segments_path.iterdir():
#         if item.is_dir() and (item / "layers").is_dir():
#             segments.append(item.name)
#
#     return segments
#



from mcp.server.fastmcp import FastMCP

from ow.workspace.mcp import (
    register_workspace_initialize,
    register_workspace_list,
    register_workspace_list_resources,
)

app = FastMCP(name="out-workspace")

_ = register_workspace_initialize(app)
_ = register_workspace_list(app)
_ = register_workspace_list_resources(app)


def main():
    """Entry point for the direct execution server."""
    app.run()


if __name__ == "__main__":
    main()



from .initialize import register_workspace_initialize
from .list import register_workspace_list, register_workspace_list_resources

__all__ = [
    "register_workspace_initialize",
    "register_workspace_list",
    "register_workspace_list_resources",
]



import typer

from pathlib import Path
from rich import print as rprint

from .options import WorkspaceOption


def get_workspace_path(workspace: WorkspaceOption) -> Path:
    """
    Checks for workspace config file in current directory or throws error.
    """
    if workspace is not None:
        # Get workspace path from name.
        from ow.workspace import WorkspaceConfig

        project_root = WorkspaceConfig.get_project_root_from_package()
        workspace_dir = project_root / "out" / workspace

    else:
        # Check for workspace config file in current directory
        workspace_dir = Path.cwd()

    config_file = workspace_dir / "config.json"

    if not config_file.exists():
        rprint(
            f"❌ [red]This is not a valid workspace folder. `{config_file}` not found.[/red]"
        )
        raise typer.Exit(code=1)

    return workspace_dir



import importlib.util
import re

from pydantic import BaseModel, field_validator, model_validator
from pathlib import Path


class WorkspaceConfig(BaseModel):
    name: str
    out_path: Path | None = None
    workspace_path: Path | None = None

    @field_validator("name", mode="before")
    @classmethod
    def normalize_and_sanitize_name(cls, v: str) -> str:
        v = v.replace(" ", "_")
        v = re.sub(r'[<>:"/\\|?*\x00-\x1F]', "", v)
        return v[:255]

    @classmethod
    def get_project_root_from_package(cls, parents_index: int = 4) -> Path:
        """Find project root based on package installation location."""
        try:
            spec = importlib.util.find_spec("ow")
            if spec and spec.origin:
                package_path = Path(spec.origin).parent
                parent_folder = package_path.parent.name
                if parent_folder == "src":
                    # Local Development
                    # package_path: /.../out-workspace/src/ow
                    # package_path.parent.parent: /.../out-workspace
                    return package_path.parent.parent
                else:
                    # PyPI Install
                    # package_path: /.../out-workspace-agent/.venv/lib/python3.13/site-packages/am
                    # package_path.parents[parents_index]: /.../out-workspace-agent
                    return package_path.parents[parents_index]
        except ImportError:
            pass
        return Path.cwd()

    @model_validator(mode="after")
    def populate_missing_paths(self) -> "WorkspaceConfig":
        if not self.out_path:
            self.out_path = self.get_project_root_from_package() / "out"

        if not self.workspace_path:
            self.workspace_path = self.out_path / self.name

        return self

    def save(self, path: Path | None = None) -> Path:
        """
        Save the configuration to a YAML file.
        If no path is given, saves to '<workspace_path>/workspace.json'.
        """
        if path is None:
            if not self.workspace_path:
                raise ValueError(
                    "workspace_path must be set to determine save location."
                )
            path = self.workspace_path / "workspace.json"

        path.parent.mkdir(parents=True, exist_ok=True)
        _ = path.write_text(self.model_dump_json(indent=2))

        return path

    @classmethod
    def load(cls: type["WorkspaceConfig"], path: Path) -> "WorkspaceConfig":
        if not path.exists():
            raise FileNotFoundError(f"Config file not found at {path}")

        return cls.model_validate_json(path.read_text())



# import shutil

# from importlib.resources import files
from pathlib import Path
from rich import print as rprint

# from am import data
from ow.workspace.config import WorkspaceConfig


class Workspace:
    """
    Base workspace methods.
    """

    def __init__(
        self,
        name: str,
        out_path: Path | None = None,
        workspace_path: Path | None = None,
    ):
        self.config: WorkspaceConfig = WorkspaceConfig(
            name=name, out_path=out_path, workspace_path=workspace_path
        )

    @property
    def name(self):
        return self.config.name

    @property
    def workspace_path(self):
        return self.config.workspace_path

    @workspace_path.setter
    def workspace_path(self, value: Path):
        self.config.workspace_path = value

    # def create_workspace_folders(self, workspace_path: Path):
    #     """
    #     Creates folders inputs and outputs within workspace.
    #     """
    #
    #     # TODO: Move `mesurements` and `meshes` into a parent `runs` folder.
    #     folders = [
    #         "materials",
    #         "measurements",
    #         "meshes",
    #         "parts",
    #         "process_maps",
    #         "segments",
    #     ]
    #     for folder in folders:
    #         resource_dir = files(data) / folder
    #         dest_dir = workspace_path / folder
    #         dest_dir.mkdir(parents=True, exist_ok=True)
    #
    #         # Copies over package resources to specific folder in workspace.
    #         for entry in resource_dir.iterdir():
    #             if entry.is_file():
    #                 dest_file = dest_dir / entry.name
    #                 with entry.open("rb") as src, open(dest_file, "wb") as dst:
    #                     shutil.copyfileobj(src, dst)

    def create_workspace(
        self,
        out_path: Path | None = None,
        force: bool | None = False,
    ) -> WorkspaceConfig:
        # Use the out_path if provided, otherwise default to package out_path.
        if out_path is None:
            out_path = self.config.out_path
            assert out_path is not None

        # Create the `out` directory if it doesn't exist.
        out_path.mkdir(exist_ok=True)

        workspace_path = out_path / self.config.name

        if workspace_path.exists() and not force:
            rprint(
                f"⚠️  [yellow]Configuration already exists at {workspace_path}[/yellow]"
            )
            rprint("Use [cyan]--force[/cyan] to overwrite, or edit the existing file.")
            raise FileExistsError("Workspace already exists")

        # self.create_workspace_folders(workspace_path)

        self.config.workspace_path = workspace_path
        workspace_config_file = self.config.save()

        rprint(f"Workspace config file saved at: {workspace_config_file}")

        return self.config



from ow.mcp.types import T, ToolError, ToolSuccess


# Convenience function to create error responses
def tool_error(message: str, code: str, **details) -> ToolError:
    """Create a standardized tool error response."""
    return ToolError(error=message, error_code=code, details=details)


# Convenience function to create success responses
def tool_success(data: T) -> ToolSuccess[T]:
    """Create a standardized tool success response."""
    return ToolSuccess(data=data)



import typer

from typing_extensions import Annotated

VerboseOption = Annotated[
    bool | None, typer.Option("--verbose", "-v", help="Enable verbose logging")
]

WorkspaceOption = Annotated[
    str | None, typer.Option("--workspace", "-w", help="Workspace to perform operation")
]



from .__main__ import app
from .development import register_mcp_development
from .install import register_mcp_install

_ = register_mcp_development(app)
_ = register_mcp_install(app)

__all__ = ["app"]



import typer

from pathlib import Path
from rich import print as rprint
from typing_extensions import Annotated
from ow.cli import WorkspaceOption


def print_list(name: str, values: list[str] | None = None):
    rprint(f"\n  {name}:")
    if values is None:
        rprint(f"  ⚠️  [yellow]No {name} found.[/yellow]")
    else:
        for index, value in enumerate(values):
            rprint(f"  {index + 1}. [cyan]{value}[/cyan]")


def register_workspace_list(app: typer.Typer):
    @app.command(name="list")
    def workspace_list(
        # meshes: Annotated[bool | None, typer.Option("--meshes", "-m", help="List meshes in workspace")] = None,
        # parts: Annotated[bool | None, typer.Option("--parts", "-p", help="List parts in workspace")] = None,
        # segments: Annotated[bool | None, typer.Option("--segments", "-s", help="List segments in workspace")] = None,
        all: Annotated[
            bool | None, typer.Option("--all", "-a", help="List all in workspace")
        ] = None,
        workspace: WorkspaceOption = None,
    ) -> None:
        """List created workspaces."""
        if workspace is not None:
            # Get workspace path from name.
            from ow.workspace import WorkspaceConfig

            project_root = WorkspaceConfig.get_project_root_from_package()
            workspace_dir = project_root / "out" / workspace

        else:
            # Check for workspace config file in current directory
            workspace_dir = Path.cwd()

        config_file = workspace_dir / "workspace.json"

        if not config_file.exists():
            # Default to showing available workspaces if no workspace_path.
            from ow.workspace.list import list_workspaces

            try:
                workspace_names = list_workspaces()
                print_list("Workspaces", workspace_names)
            except:
                rprint("⚠️  [yellow]Unable to list workspaces[/yellow]")
                _ = typer.Exit()
        else:
            rprint(f"Workspace: {workspace}")

            # if all or meshes:
            #     from ow.workspace.list import list_workspace_meshes
            #
            #     try:
            #         workspace_meshes = list_workspace_meshes(workspace_dir)
            #         print_list("Meshes", workspace_meshes)
            #     except:
            #         rprint("⚠️  [yellow]Unable to list meshes[/yellow]")
            #         _ = typer.Exit()
            #
            # if all or parts:
            #     from am.workspace.list import list_workspace_parts
            #
            #     try:
            #         workspace_parts = list_workspace_parts(workspace_dir)
            #         print_list("Parts", workspace_parts)
            #     except:
            #         rprint("⚠️  [yellow]Unable to list parts[/yellow]")
            #         _ = typer.Exit()
            #
            # if all or segments:
            #     from am.workspace.list import list_workspace_segments
            #
            #     try:
            #         workspace_segments = list_workspace_segments(workspace_dir)
            #         print_list("Segments", workspace_segments)
            #     except:
            #         rprint("⚠️  [yellow]Unable to list segments[/yellow]")
            #         _ = typer.Exit()

    _ = app.command(name="ls")(workspace_list)

    _ = workspace_list
